@@ -52,6 +52,11 @@
 #include <limits.h>
 #endif
 
+/* PATH_MAX is undefined on the hurd */
+#if !defined(PATH_MAX) && defined(_POSIX_PATH_MAX)
+#define PATH_MAX _POSIX_PATH_MAX
+#endif
+
 #include <pwd.h>
 #include <grp.h>
 
@@ -66,6 +71,20 @@
 # define MB_LEN_MAX 16
 #endif
 
+#ifdef HAVE_FGETS_UNLOCKED
+# ifdef fgets
+#  undef fgets
+# endif
+# define fgets fgets_unlocked
+#endif
+
+#ifdef HAVE_FGETC_UNLOCKED
+# ifdef fgetc
+#  undef fgetc
+# endif
+# define fgetc fgetc_unlocked
+#endif
+
 /* nifty trick I stole from ELM 2.5alpha. */
 #ifdef MAIN_C
 #define WHERE 
@@ -80,23 +99,24 @@
 #include "mutt_regex.h"
 
 /* flags for mutt_enter_string() */
-#define  M_ALIAS   1      /* do alias "completion" by calling up the alias-menu */
-#define  M_FILE    (1<<1) /* do file completion */
-#define  M_EFILE   (1<<2) /* do file completion, plus incoming folders */
-#define  M_CMD     (1<<3) /* do completion on previous word */
-#define  M_PASS    (1<<4) /* password mode (no echo) */
-#define  M_CLEAR   (1<<5) /* clear input if printable character is pressed */
-#define  M_COMMAND (1<<6) /* do command completion */
-#define  M_PATTERN (1<<7) /* pattern mode - only used for history classes */
+#define  MUTT_ALIAS   1      /* do alias "completion" by calling up the alias-menu */
+#define  MUTT_FILE    (1<<1) /* do file completion */
+#define  MUTT_EFILE   (1<<2) /* do file completion, plus incoming folders */
+#define  MUTT_CMD     (1<<3) /* do completion on previous word */
+#define  MUTT_PASS    (1<<4) /* password mode (no echo) */
+#define  MUTT_CLEAR   (1<<5) /* clear input if printable character is pressed */
+#define  MUTT_COMMAND (1<<6) /* do command completion */
+#define  MUTT_PATTERN (1<<7) /* pattern mode - only used for history classes */
+#define  MUTT_LABEL   (1<<8) /* do label completion */
 
 /* flags for mutt_get_token() */
-#define M_TOKEN_EQUAL		1	/* treat '=' as a special */
-#define M_TOKEN_CONDENSE	(1<<1)	/* ^(char) to control chars (macros) */
-#define M_TOKEN_SPACE		(1<<2)  /* don't treat whitespace as a term */
-#define M_TOKEN_QUOTE		(1<<3)	/* don't interpret quotes */
-#define M_TOKEN_PATTERN		(1<<4)	/* !)|~ are terms (for patterns) */
-#define M_TOKEN_COMMENT		(1<<5)	/* don't reap comments */
-#define M_TOKEN_SEMICOLON	(1<<6)	/* don't treat ; as special */
+#define MUTT_TOKEN_EQUAL      1       /* treat '=' as a special */
+#define MUTT_TOKEN_CONDENSE   (1<<1)  /* ^(char) to control chars (macros) */
+#define MUTT_TOKEN_SPACE      (1<<2)  /* don't treat whitespace as a term */
+#define MUTT_TOKEN_QUOTE      (1<<3)  /* don't interpret quotes */
+#define MUTT_TOKEN_PATTERN    (1<<4)  /* !)|~ are terms (for patterns) */
+#define MUTT_TOKEN_COMMENT    (1<<5)  /* don't reap comments */
+#define MUTT_TOKEN_SEMICOLON  (1<<6)  /* don't treat ; as special */
 
 typedef struct
 {
@@ -113,148 +133,159 @@ typedef struct
 } event_t;
 
 /* flags for _mutt_system() */
-#define M_DETACH_PROCESS	1	/* detach subprocess from group */
+#define MUTT_DETACH_PROCESS	1	/* detach subprocess from group */
 
 /* flags for mutt_FormatString() */
 typedef enum
 {
-  M_FORMAT_FORCESUBJ	= (1<<0), /* print the subject even if unchanged */
-  M_FORMAT_TREE		= (1<<1), /* draw the thread tree */
-  M_FORMAT_MAKEPRINT	= (1<<2), /* make sure that all chars are printable */
-  M_FORMAT_OPTIONAL	= (1<<3),
-  M_FORMAT_STAT_FILE	= (1<<4), /* used by mutt_attach_fmt */
-  M_FORMAT_ARROWCURSOR	= (1<<5), /* reserve space for arrow_cursor */
-  M_FORMAT_INDEX	= (1<<6), /* this is a main index entry */
-  M_FORMAT_NOFILTER	= (1<<7)  /* do not allow filtering on this pass */
+  MUTT_FORMAT_FORCESUBJ   = (1<<0), /* print the subject even if unchanged */
+  MUTT_FORMAT_TREE        = (1<<1), /* draw the thread tree */
+  MUTT_FORMAT_MAKEPRINT   = (1<<2), /* make sure that all chars are printable */
+  MUTT_FORMAT_OPTIONAL    = (1<<3),
+  MUTT_FORMAT_STAT_FILE   = (1<<4), /* used by mutt_attach_fmt */
+  MUTT_FORMAT_ARROWCURSOR = (1<<5), /* reserve space for arrow_cursor */
+  MUTT_FORMAT_INDEX       = (1<<6), /* this is a main index entry */
+  MUTT_FORMAT_NOFILTER    = (1<<7)  /* do not allow filtering on this pass */
 } format_flag;
 
 /* types for mutt_add_hook() */
-#define M_FOLDERHOOK	1
-#define M_MBOXHOOK	(1<<1)
-#define M_SENDHOOK	(1<<2)
-#define M_FCCHOOK	(1<<3)
-#define M_SAVEHOOK	(1<<4)
-#define M_CHARSETHOOK	(1<<5)
-#define M_ICONVHOOK 	(1<<6)
-#define M_MESSAGEHOOK	(1<<7)
-#define M_CRYPTHOOK	(1<<8)
-#define M_ACCOUNTHOOK	(1<<9)
-#define M_REPLYHOOK	(1<<10)
-#define M_SEND2HOOK     (1<<11)
+#define MUTT_FOLDERHOOK  1
+#define MUTT_MBOXHOOK    (1<<1)
+#define MUTT_SENDHOOK    (1<<2)
+#define MUTT_FCCHOOK     (1<<3)
+#define MUTT_SAVEHOOK    (1<<4)
+#define MUTT_CHARSETHOOK (1<<5)
+#define MUTT_ICONVHOOK   (1<<6)
+#define MUTT_MESSAGEHOOK (1<<7)
+#define MUTT_CRYPTHOOK   (1<<8)
+#define MUTT_ACCOUNTHOOK (1<<9)
+#define MUTT_REPLYHOOK   (1<<10)
+#define MUTT_SEND2HOOK   (1<<11)
+#ifdef USE_COMPRESSED
+#define MUTT_OPENHOOK    (1<<12)
+#define MUTT_APPENDHOOK  (1<<13)
+#define MUTT_CLOSEHOOK   (1<<14)
+#endif
 
 /* tree characters for linearize_tree and print_enriched_string */
-#define M_TREE_LLCORNER		1
-#define M_TREE_ULCORNER		2
-#define M_TREE_LTEE		3
-#define M_TREE_HLINE		4
-#define M_TREE_VLINE		5
-#define M_TREE_SPACE		6
-#define M_TREE_RARROW		7
-#define M_TREE_STAR		8
-#define M_TREE_HIDDEN		9
-#define M_TREE_EQUALS		10
-#define M_TREE_TTEE		11
-#define M_TREE_BTEE		12
-#define M_TREE_MISSING		13
-#define M_TREE_MAX		14
+#define MUTT_TREE_LLCORNER      1
+#define MUTT_TREE_ULCORNER      2
+#define MUTT_TREE_LTEE          3
+#define MUTT_TREE_HLINE         4
+#define MUTT_TREE_VLINE         5
+#define MUTT_TREE_SPACE         6
+#define MUTT_TREE_RARROW        7
+#define MUTT_TREE_STAR          8
+#define MUTT_TREE_HIDDEN        9
+#define MUTT_TREE_EQUALS        10
+#define MUTT_TREE_TTEE          11
+#define MUTT_TREE_BTEE          12
+#define MUTT_TREE_MISSING       13
+#define MUTT_TREE_MAX           14
 
-#define M_THREAD_COLLAPSE	(1<<0)
-#define M_THREAD_UNCOLLAPSE	(1<<1)
-#define M_THREAD_GET_HIDDEN	(1<<2)
-#define M_THREAD_UNREAD		(1<<3)
-#define M_THREAD_NEXT_UNREAD	(1<<4)
+#define MUTT_THREAD_COLLAPSE    (1<<0)
+#define MUTT_THREAD_UNCOLLAPSE  (1<<1)
+#define MUTT_THREAD_GET_HIDDEN  (1<<2)
+#define MUTT_THREAD_UNREAD      (1<<3)
+#define MUTT_THREAD_NEXT_UNREAD (1<<4)
 
 enum
 {
   /* modes for mutt_view_attachment() */
-  M_REGULAR = 1,
-  M_MAILCAP,
-  M_AS_TEXT,
+  MUTT_REGULAR = 1,
+  MUTT_MAILCAP,
+  MUTT_AS_TEXT,
 
   /* action codes used by mutt_set_flag() and mutt_pattern_function() */
-  M_ALL,
-  M_NONE,
-  M_NEW,
-  M_OLD,
-  M_REPLIED,
-  M_READ,
-  M_UNREAD,
-  M_DELETE,
-  M_UNDELETE,
-  M_DELETED,
-  M_FLAG,
-  M_TAG,
-  M_UNTAG,
-  M_LIMIT,
-  M_EXPIRED,
-  M_SUPERSEDED,
+  MUTT_ALL,
+  MUTT_NONE,
+  MUTT_NEW,
+  MUTT_OLD,
+  MUTT_REPLIED,
+  MUTT_READ,
+  MUTT_UNREAD,
+  MUTT_DELETE,
+  MUTT_UNDELETE,
+  MUTT_PURGE,
+  MUTT_DELETED,
+  MUTT_FLAG,
+  MUTT_TAG,
+  MUTT_UNTAG,
+  MUTT_LIMIT,
+  MUTT_EXPIRED,
+  MUTT_SUPERSEDED,
+  MUTT_TRASH,
 
   /* actions for mutt_pattern_comp/mutt_pattern_exec */
-  M_AND,
-  M_OR,
-  M_THREAD,
-  M_TO,
-  M_CC,
-  M_COLLAPSED,
-  M_SUBJECT,
-  M_FROM,
-  M_DATE,
-  M_DATE_RECEIVED,
-  M_DUPLICATED,
-  M_UNREFERENCED,
-  M_ID,
-  M_BODY,
-  M_HEADER,
-  M_HORMEL,
-  M_WHOLE_MSG,
-  M_SENDER,
-  M_MESSAGE,
-  M_SCORE,
-  M_SIZE,
-  M_REFERENCE,
-  M_RECIPIENT,
-  M_LIST,
-  M_SUBSCRIBED_LIST,
-  M_PERSONAL_RECIP,
-  M_PERSONAL_FROM,
-  M_ADDRESS,
-  M_CRYPT_SIGN,
-  M_CRYPT_VERIFIED,
-  M_CRYPT_ENCRYPT,
-  M_PGP_KEY,
-  M_XLABEL,
-  M_MIMEATTACH,
+  MUTT_AND,
+  MUTT_OR,
+  MUTT_THREAD,
+  MUTT_PARENT,
+  MUTT_CHILDREN,
+  MUTT_TO,
+  MUTT_CC,
+  MUTT_COLLAPSED,
+  MUTT_SUBJECT,
+  MUTT_FROM,
+  MUTT_DATE,
+  MUTT_DATE_RECEIVED,
+  MUTT_DUPLICATED,
+  MUTT_UNREFERENCED,
+  MUTT_ID,
+  MUTT_BODY,
+  MUTT_HEADER,
+  MUTT_HORMEL,
+  MUTT_WHOLE_MSG,
+  MUTT_SENDER,
+  MUTT_MESSAGE,
+  MUTT_SCORE,
+  MUTT_SIZE,
+  MUTT_REFERENCE,
+  MUTT_RECIPIENT,
+  MUTT_LIST,
+  MUTT_SUBSCRIBED_LIST,
+  MUTT_PERSONAL_RECIP,
+  MUTT_PERSONAL_FROM,
+  MUTT_ADDRESS,
+  MUTT_CRYPT_SIGN,
+  MUTT_CRYPT_VERIFIED,
+  MUTT_CRYPT_ENCRYPT,
+  MUTT_PGP_KEY,
+  MUTT_XLABEL,
+  MUTT_MIMEATTACH,
+  MUTT_MIMETYPE,
   
   /* Options for Mailcap lookup */
-  M_EDIT,
-  M_COMPOSE,
-  M_PRINT,
-  M_AUTOVIEW,
+  MUTT_EDIT,
+  MUTT_COMPOSE,
+  MUTT_PRINT,
+  MUTT_AUTOVIEW,
 
   /* options for socket code */
-  M_NEW_SOCKET,
+  MUTT_NEW_SOCKET,
 #ifdef USE_SSL_OPENSSL
-  M_NEW_SSL_SOCKET,
+  MUTT_NEW_SSL_SOCKET,
 #endif
 
   /* Options for mutt_save_attachment */
-  M_SAVE_APPEND,
-  M_SAVE_OVERWRITE
+  MUTT_SAVE_APPEND,
+  MUTT_SAVE_OVERWRITE
 };
 
 /* possible arguments to set_quadoption() */
 enum
 {
-  M_NO,
-  M_YES,
-  M_ASKNO,
-  M_ASKYES
+  MUTT_NO,
+  MUTT_YES,
+  MUTT_ASKNO,
+  MUTT_ASKYES
 };
 
 /* quad-option vars */
 enum
 {
   OPT_ABORT,
+  OPT_ABORTNOATTACH,
   OPT_BOUNCE,
   OPT_COPY,
   OPT_DELETE,
@@ -300,16 +331,16 @@ enum
 #define SENDDRAFTFILE		(1<<11)   /* Used by the -H flag */
 
 /* flags for mutt_compose_menu() */
-#define M_COMPOSE_NOFREEHEADER (1<<0)
+#define MUTT_COMPOSE_NOFREEHEADER (1<<0)
 
 /* flags to _mutt_select_file() */
-#define M_SEL_BUFFY	(1<<0)
-#define M_SEL_MULTI	(1<<1)
-#define M_SEL_FOLDER	(1<<2)
+#define MUTT_SEL_BUFFY  (1<<0)
+#define MUTT_SEL_MULTI  (1<<1)
+#define MUTT_SEL_FOLDER (1<<2)
 
 /* flags for parse_spam_list */
-#define M_SPAM          1
-#define M_NOSPAM        2
+#define MUTT_SPAM          1
+#define MUTT_NOSPAM        2
 
 /* boolean vars */
 enum
@@ -326,7 +357,9 @@ enum
   OPTBEEP,
   OPTBEEPNEW,
   OPTBOUNCEDELIVERED,
+  OPTCHANGEFOLDERNEXT,
   OPTBRAILLEFRIENDLY,
+  OPTBROWSERABBRMAILBOXES,
   OPTCHECKMBOXSIZE,
   OPTCHECKNEW,
   OPTCOLLAPSEUNREAD,
@@ -340,18 +373,20 @@ enum
   OPTENVFROM,
   OPTFASTREPLY,
   OPTFCCCLEAR,
+  OPTFLAGSAFE,
   OPTFOLLOWUPTO,
   OPTFORCENAME,
   OPTFORWDECODE,
   OPTFORWQUOTE,
 #ifdef USE_HCACHE
   OPTHCACHEVERIFY,
-#if defined(HAVE_QDBM) || defined(HAVE_TC)
+#if defined(HAVE_QDBM) || defined(HAVE_TC) || defined(HAVE_KC)
   OPTHCACHECOMPRESS,
 #endif /* HAVE_QDBM */
 #endif
   OPTHDRS,
   OPTHEADER,
+  OPTHEADERCOLORPARTIAL,
   OPTHELP,
   OPTHIDDENHOST,
   OPTHIDELIMITED,
@@ -359,6 +394,7 @@ enum
   OPTHIDETHREADSUBJECT,
   OPTHIDETOPLIMITED,
   OPTHIDETOPMISSING,
+  OPTHISTREMOVEDUPS,
   OPTHONORDISP,
   OPTIGNORELWS,
   OPTIGNORELISTREPLYTO,
@@ -382,12 +418,16 @@ enum
   OPTSSLFORCETLS,
   OPTSSLVERIFYDATES,
   OPTSSLVERIFYHOST,
+# if defined(USE_SSL_OPENSSL) && defined(HAVE_SSL_PARTIAL_CHAIN)
+  OPTSSLVERIFYPARTIAL,
+# endif /* USE_SSL_OPENSSL */
 #endif /* defined(USE_SSL) */
   OPTIMPLICITAUTOVIEW,
   OPTINCLUDEONLYFIRST,
   OPTKEEPFLAGGED,
   OPTMAILCAPSANITIZE,
   OPTMAILCHECKRECENT,
+  OPTMAILCHECKSTATS,
   OPTMAILDIRTRASH,
   OPTMAILDIRCHECKCUR,
   OPTMARKERS,
@@ -401,6 +441,7 @@ enum
   OPTMETOO,
   OPTMHPURGE,
   OPTMIMEFORWDECODE,
+  OPTMIMETYPEQUERYFIRST,
   OPTNARROWTREE,
   OPTPAGERSTOP,
   OPTPIPEDECODE,
@@ -428,6 +469,13 @@ enum
   OPTSAVEEMPTY,
   OPTSAVENAME,
   OPTSCORE,
+#ifdef USE_SIDEBAR
+  OPTSIDEBAR,
+  OPTSIDEBARFOLDERINDENT,
+  OPTSIDEBARNEWMAILONLY,
+  OPTSIDEBARNEXTNEWWRAP,
+  OPTSIDEBARSHORTPATH,
+#endif
   OPTSIGDASHES,
   OPTSIGONTOP,
   OPTSORTRE,
@@ -441,11 +489,12 @@ enum
   OPTTILDE,
   OPTTSENABLED,
   OPTUNCOLLAPSEJUMP,
+  OPTUNCOLLAPSENEW,
   OPTUSE8BITMIME,
   OPTUSEDOMAIN,
   OPTUSEFROM,
   OPTUSEGPGAGENT,
-#ifdef HAVE_LIBIDN
+#if defined(HAVE_LIBIDN) || defined(HAVE_LIBIDN2)
   OPTIDNDECODE,
   OPTIDNENCODE,
 #endif
@@ -475,16 +524,19 @@ enum
   OPTCRYPTREPLYSIGNENCRYPTED,
   OPTCRYPTTIMESTAMP,
   OPTSMIMEISDEFAULT,
+  OPTSMIMESELFENCRYPT,
   OPTASKCERTLABEL,
   OPTSDEFAULTDECRYPTKEY,
   OPTPGPIGNORESUB,
   OPTPGPCHECKEXIT,
+  OPTPGPCHECKGPGDECRYPTSTATUSFD,
   OPTPGPLONGIDS,
   OPTPGPAUTODEC,
 #if 0
   OPTPGPENCRYPTSELF,
 #endif
   OPTPGPRETAINABLESIG,
+  OPTPGPSELFENCRYPT,
   OPTPGPSTRICTENC,
   OPTFORWDECRYPT,
   OPTPGPSHOWUNUSABLE,
@@ -497,7 +549,6 @@ enum
   OPTFORCEREFRESH,	/* (pseudo) refresh even during macros */
   OPTLOCALES,		/* (pseudo) set if user has valid locale definition */
   OPTNOCURSES,		/* (pseudo) when sending in batch mode */
-  OPTNEEDREDRAW,	/* (pseudo) to notify caller of a submenu */
   OPTSEARCHREVERSE,	/* (pseudo) used by ci_search_command */
   OPTMSGERR,		/* (pseudo) used by mutt_error/mutt_message */
   OPTSEARCHINVALID,	/* (pseudo) used to invalidate the search pat */
@@ -506,8 +557,6 @@ enum
   OPTNEEDRESORT,	/* (pseudo) used to force a re-sort */
   OPTRESORTINIT,	/* (pseudo) used to force the next resort to be from scratch */
   OPTVIEWATTACH,	/* (pseudo) signals that we are viewing attachments */
-  OPTFORCEREDRAWINDEX,	/* (pseudo) used to force a redraw in the main index */
-  OPTFORCEREDRAWPAGER,	/* (pseudo) used to force a redraw in the pager */
   OPTSORTSUBTHREADS,	/* (pseudo) used when $sort_aux changes */
   OPTNEEDRESCORE,	/* (pseudo) set when the `score' command is used */
   OPTATTACHMSG,		/* (pseudo) used by attach-message */
@@ -547,20 +596,20 @@ typedef struct rx_list_t
   struct rx_list_t *next;
 } RX_LIST;
 
-typedef struct spam_list_t
+typedef struct replace_list_t
 {
   REGEXP *rx;
   int     nmatch;
   char   *template;
-  struct spam_list_t *next;
-} SPAM_LIST;
+  struct replace_list_t *next;
+} REPLACE_LIST;
 
 #define mutt_new_list() safe_calloc (1, sizeof (LIST))
 #define mutt_new_rx_list() safe_calloc (1, sizeof (RX_LIST))
-#define mutt_new_spam_list() safe_calloc (1, sizeof (SPAM_LIST))
+#define mutt_new_replace_list() safe_calloc (1, sizeof (REPLACE_LIST))
 void mutt_free_list (LIST **);
 void mutt_free_rx_list (RX_LIST **);
-void mutt_free_spam_list (SPAM_LIST **);
+void mutt_free_replace_list (REPLACE_LIST **);
 LIST *mutt_copy_list (LIST *);
 int mutt_matches_ignore (const char *, LIST *);
 
@@ -596,6 +645,7 @@ typedef struct envelope
   char *list_post;		/* this stores a mailto URL, or nothing */
   char *subject;
   char *real_subj;		/* offset of the real subject */
+  char *disp_subj;		/* display subject (modified copy of subject) */
   char *message_id;
   char *supersedes;
   char *date;
@@ -621,6 +671,7 @@ typedef struct content
 {
   long hibin;              /* 8-bit characters */
   long lobin;              /* unprintable 7-bit chars (eg., control chars) */
+  long nulbin;             /* null characters (0x0) */
   long crlf;		   /* '\r' and '\n' characters */
   long ascii;              /* number of ascii chars */
   long linemax;            /* length of the longest line in the file */
@@ -720,6 +771,7 @@ typedef struct header
   unsigned int flagged : 1; 		/* marked important? */
   unsigned int tagged : 1;
   unsigned int deleted : 1;
+  unsigned int purge : 1;               /* skip trash folder when deleting */
   unsigned int changed : 1;
   unsigned int attach_del : 1; 		/* has an attachment marked for deletion */
   unsigned int old : 1;
@@ -736,6 +788,7 @@ typedef struct header
 					 * This flag is used by the maildir_trash
 					 * option.
 					 */
+  unsigned int xlabel_changed : 1;	/* editable - used for syncing */
   
   /* timezone of the sender of this message */
   unsigned int zhours : 5;
@@ -811,10 +864,10 @@ struct mutt_thread
 
 
 /* flag to mutt_pattern_comp() */
-#define M_FULL_MSG	(1<<0)	/* enable body and header matching */
+#define MUTT_FULL_MSG	(1<<0)	/* enable body and header matching */
 
 typedef enum {
-  M_MATCH_FULL_ADDRESS = 1
+  MUTT_MATCH_FULL_ADDRESS = 1
 } pattern_exec_flag;
 
 typedef struct group_t
@@ -838,6 +891,7 @@ typedef struct pattern_t
   unsigned int stringmatch : 1;
   unsigned int groupmatch : 1;
   unsigned int ign_case : 1;		/* ignore case for local stringmatch searches */
+  unsigned int isalias : 1;
   int min;
   int max;
   struct pattern_t *next;
@@ -850,28 +904,73 @@ typedef struct pattern_t
   } p;
 } pattern_t;
 
+/* This is used when a message is repeatedly pattern matched against.
+ * e.g. for color, scoring, hooks.  It caches a few of the potentially slow
+ * operations.
+ * Each entry has a value of 0 = unset, 1 = false, 2 = true
+ */
+typedef struct
+{
+  int list_all;          /* ^~l */
+  int list_one;          /*  ~l */
+  int sub_all;           /* ^~u */
+  int sub_one;           /*  ~u */
+  int pers_recip_all;    /* ^~p */
+  int pers_recip_one;    /*  ~p */
+  int pers_from_all;     /* ^~P */
+  int pers_from_one;     /*  ~P */
+} pattern_cache_t;
+
 /* ACL Rights */
 enum
 {
-  M_ACL_LOOKUP = 0,
-  M_ACL_READ,
-  M_ACL_SEEN,
-  M_ACL_WRITE,
-  M_ACL_INSERT,
-  M_ACL_POST,
-  M_ACL_CREATE,
-  M_ACL_DELMX,
-  M_ACL_DELETE,
-  M_ACL_EXPUNGE,
-  M_ACL_ADMIN,
+  MUTT_ACL_LOOKUP = 0,
+  MUTT_ACL_READ,
+  MUTT_ACL_SEEN,
+  MUTT_ACL_WRITE,
+  MUTT_ACL_INSERT,
+  MUTT_ACL_POST,
+  MUTT_ACL_CREATE,
+  MUTT_ACL_DELMX,
+  MUTT_ACL_DELETE,
+  MUTT_ACL_EXPUNGE,
+  MUTT_ACL_ADMIN,
 
   RIGHTSMAX
 };
 
+struct _context;
+struct _message;
+
+/*
+ * struct mx_ops - a structure to store operations on a mailbox
+ * The following operations are mandatory:
+ *  - open
+ *  - close
+ *  - check
+ *
+ * Optional operations
+ *  - open_new_msg
+ */
+struct mx_ops
+{
+  int (*open) (struct _context *);
+  int (*open_append) (struct _context *, int flags);
+  int (*close) (struct _context *);
+  int (*check) (struct _context *ctx, int *index_hint);
+  int (*sync) (struct _context *ctx, int *index_hint);
+  int (*open_msg) (struct _context *, struct _message *, int msgno);
+  int (*close_msg) (struct _context *, struct _message *);
+  int (*commit_msg) (struct _context *, struct _message *);
+  int (*open_new_msg) (struct _message *, struct _context *, HEADER *);
+};
+
 typedef struct _context
 {
   char *path;
+  char *realpath;               /* used for buffy comparison and the sidebar */
   FILE *fp;
+  time_t atime;
   time_t mtime;
   off_t size;
   off_t vsize;
@@ -883,6 +982,7 @@ typedef struct _context
   HASH *id_hash;		/* hash table by msg id */
   HASH *subj_hash;		/* hash table by subject */
   HASH *thread_hash;		/* hash table for threading */
+  HASH *label_hash;             /* hash table for x-labels */
   int *v2r;			/* mapping from virtual to real msgno */
   int hdrmax;			/* number of pointers in hdrs */
   int msgcount;			/* number of messages in the mailbox */
@@ -906,10 +1006,15 @@ typedef struct _context
   unsigned int quiet : 1;	/* inhibit status messages? */
   unsigned int collapsed : 1;   /* are all threads collapsed? */
   unsigned int closing : 1;	/* mailbox is being closed */
+  unsigned int peekonly : 1;	/* just taking a glance, revert atime */
+
+#ifdef USE_COMPRESSED
+  void *compress_info;		/* compressed mbox module private data */
+#endif /* USE_COMPRESSED */
 
   /* driver hooks */
   void *data;			/* driver specific data */
-  int (*mx_close)(struct _context *);
+  struct mx_ops *mx_ops;
 } CONTEXT;
 
 typedef struct
@@ -933,17 +1038,17 @@ typedef struct
 } ENTER_STATE;
 
 /* flags for the STATE struct */
-#define M_DISPLAY	(1<<0) /* output is displayed to the user */
-#define M_VERIFY	(1<<1) /* perform signature verification */
-#define M_PENDINGPREFIX (1<<2) /* prefix to write, but character must follow */
-#define M_WEED          (1<<3) /* weed headers even when not in display mode */
-#define M_CHARCONV	(1<<4) /* Do character set conversions */
-#define M_PRINTING	(1<<5) /* are we printing? - M_DISPLAY "light" */
-#define M_REPLYING	(1<<6) /* are we replying? */
-#define M_FIRSTDONE	(1<<7) /* the first attachment has been done */
+#define MUTT_DISPLAY       (1<<0) /* output is displayed to the user */
+#define MUTT_VERIFY        (1<<1) /* perform signature verification */
+#define MUTT_PENDINGPREFIX (1<<2) /* prefix to write, but character must follow */
+#define MUTT_WEED          (1<<3) /* weed headers even when not in display mode */
+#define MUTT_CHARCONV      (1<<4) /* Do character set conversions */
+#define MUTT_PRINTING      (1<<5) /* are we printing? - MUTT_DISPLAY "light" */
+#define MUTT_REPLYING      (1<<6) /* are we replying? */
+#define MUTT_FIRSTDONE     (1<<7) /* the first attachment has been done */
 
-#define state_set_prefix(s) ((s)->flags |= M_PENDINGPREFIX)
-#define state_reset_prefix(s) ((s)->flags &= ~M_PENDINGPREFIX)
+#define state_set_prefix(s) ((s)->flags |= MUTT_PENDINGPREFIX)
+#define state_reset_prefix(s) ((s)->flags &= ~MUTT_PENDINGPREFIX)
 #define state_puts(x,y) fputs(x,(y)->fpout)
 #define state_putc(x,y) fputc(x,(y)->fpout)
 
@@ -963,7 +1068,18 @@ typedef struct
   regex_t minor_rx;
 } ATTACH_MATCH;
 
-#define M_PARTS_TOPLEVEL	(1<<0)	/* is the top-level part */
+/* multibyte character table.
+ * Allows for direct access to the individual multibyte characters in a
+ * string.  This is used for the Tochars and StChars option types. */
+typedef struct
+{
+  int len;               /* number of characters */
+  char **chars;          /* the array of multibyte character strings */
+  char *segmented_str;   /* each chars entry points inside this string */
+  char *orig_str;
+} mbchar_table;
+
+#define MUTT_PARTS_TOPLEVEL	(1<<0)	/* is the top-level part */
 
 #include "ascii.h"
 #include "protos.h"
