@@ -40,6 +40,25 @@
 #define HC_FEXT		"hcache"	/* extension for hcache as POP lacks paths */
 #endif
 
+/**
+ * cache_id - Make a message-cache-compatible id
+ * @param id POP message id
+ * @retval ptr Sanitised string
+ *
+ * The POP message id may contain '/' and other awkward characters.
+ *
+ * @note This function returns a pointer to a static buffer.
+ */
+static const char *cache_id(const char *id)
+{
+  static char clean[SHORT_STRING];
+
+  strfcpy (clean, id, sizeof(clean));
+  mutt_sanitize_filename (clean, 1);
+
+  return clean;
+}
+
 /* write line to file */
 static int fetch_message (char *line, void *file)
 {
@@ -98,7 +117,7 @@ static int pop_read_header (POP_DATA *pop_data, HEADER *h)
 	pop_data->cmd_top = 0;
 
 	dprint (1, (debugfile, "pop_read_header: unset TOP capability\n"));
-	snprintf (pop_data->err_msg, sizeof (pop_data->err_msg),
+	snprintf (pop_data->err_msg, sizeof (pop_data->err_msg), "%s",
 		_("Command TOP is not supported by server."));
       }
     }
@@ -152,6 +171,10 @@ static int fetch_uidl (char *line, void *data)
       endp++;
   memmove(line, endp, strlen(endp) + 1);
 
+  /* uid must be at least be 1 byte */
+  if (strlen(line) == 0)
+    return -1;
+
   for (i = 0; i < ctx->msgcount; i++)
     if (!mutt_strcmp (line, ctx->hdrs[i]->data))
       break;
@@ -201,7 +224,7 @@ static int msg_cache_check (const char *id, body_cache_t *bcache, void *data)
   /* message not found in context -> remove it from cache
    * return the result of bcache, so we stop upon its first error
    */
-  return mutt_bcache_del (bcache, id);
+  return mutt_bcache_del (bcache, cache_id (id));
 }
 
 #ifdef USE_HCACHE
@@ -272,14 +295,14 @@ static int pop_fetch_headers (CONTEXT *ctx)
       pop_data->cmd_uidl = 0;
 
       dprint (1, (debugfile, "pop_fetch_headers: unset UIDL capability\n"));
-      snprintf (pop_data->err_msg, sizeof (pop_data->err_msg),
+      snprintf (pop_data->err_msg, sizeof (pop_data->err_msg), "%s",
 	      _("Command UIDL is not supported by server."));
     }
   }
 
   if (!ctx->quiet)
     mutt_progress_init (&progress, _("Fetching message headers..."),
-                        M_PROGRESS_MSG, ReadInc, new_count - old_count);
+                        MUTT_PROGRESS_MSG, ReadInc, new_count - old_count);
 
   if (ret == 0)
   {
@@ -333,10 +356,10 @@ static int pop_fetch_headers (CONTEXT *ctx)
 #if USE_HCACHE
       else
       {
-	mutt_hcache_store (hc, ctx->hdrs[i]->data, ctx->hdrs[i], 0, strlen, M_GENERATE_UIDVALIDITY);
+	mutt_hcache_store (hc, ctx->hdrs[i]->data, ctx->hdrs[i], 0, strlen, MUTT_GENERATE_UIDVALIDITY);
       }
 
-      FREE(&data);
+      mutt_hcache_free (&data);
 #endif
 
       /*
@@ -351,7 +374,7 @@ static int pop_fetch_headers (CONTEXT *ctx)
        *        - if we also have a body: read
        *        - if we don't have a body: new
        */
-      bcached = mutt_bcache_exists (pop_data->bcache, ctx->hdrs[i]->data) == 0;
+      bcached = mutt_bcache_exists (pop_data->bcache, cache_id (ctx->hdrs[i]->data)) == 0;
       ctx->hdrs[i]->old = 0;
       ctx->hdrs[i]->read = 0;
       if (hcached)
@@ -397,7 +420,7 @@ static int pop_fetch_headers (CONTEXT *ctx)
 }
 
 /* open POP mailbox - fetch only headers */
-int pop_open_mailbox (CONTEXT *ctx)
+static int pop_open_mailbox (CONTEXT *ctx)
 {
   int ret;
   char buf[LONG_STRING];
@@ -421,12 +444,13 @@ int pop_open_mailbox (CONTEXT *ctx)
     return -1;
 
   FREE (&ctx->path);
+  FREE (&ctx->realpath);
   ctx->path = safe_strdup (buf);
+  ctx->realpath = safe_strdup (ctx->path);
 
   pop_data = safe_calloc (1, sizeof (POP_DATA));
   pop_data->conn = conn;
   ctx->data = pop_data;
-  ctx->mx_close = pop_close_mailbox;
 
   if (pop_open_connection (pop_data) < 0)
     return -1;
@@ -436,12 +460,12 @@ int pop_open_mailbox (CONTEXT *ctx)
 
   /* init (hard-coded) ACL rights */
   memset (ctx->rights, 0, sizeof (ctx->rights));
-  mutt_bit_set (ctx->rights, M_ACL_SEEN);
-  mutt_bit_set (ctx->rights, M_ACL_DELETE);
+  mutt_bit_set (ctx->rights, MUTT_ACL_SEEN);
+  mutt_bit_set (ctx->rights, MUTT_ACL_DELETE);
 #if USE_HCACHE
   /* flags are managed using header cache, so it only makes sense to
    * enable them in that case */
-  mutt_bit_set (ctx->rights, M_ACL_WRITE);
+  mutt_bit_set (ctx->rights, MUTT_ACL_WRITE);
 #endif
 
   FOREVER
@@ -513,7 +537,7 @@ int pop_close_mailbox (CONTEXT *ctx)
 }
 
 /* fetch message from POP server */
-int pop_fetch_message (MESSAGE* msg, CONTEXT* ctx, int msgno)
+static int pop_fetch_message (CONTEXT* ctx, MESSAGE* msg, int msgno)
 {
   int ret;
   void *uidl;
@@ -526,7 +550,7 @@ int pop_fetch_message (MESSAGE* msg, CONTEXT* ctx, int msgno)
   unsigned short bcache = 1;
 
   /* see if we already have the message in body cache */
-  if ((msg->fp = mutt_bcache_get (pop_data->bcache, h->data)))
+  if ((msg->fp = mutt_bcache_get (pop_data->bcache, cache_id (h->data))))
     return 0;
 
   /*
@@ -570,10 +594,10 @@ int pop_fetch_message (MESSAGE* msg, CONTEXT* ctx, int msgno)
     }
 
     mutt_progress_init (&progressbar, _("Fetching message..."),
-			M_PROGRESS_SIZE, NetInc, h->content->length + h->content->offset - 1);
+			MUTT_PROGRESS_SIZE, NetInc, h->content->length + h->content->offset - 1);
 
     /* see if we can put in body cache; use our cache as fallback */
-    if (!(msg->fp = mutt_bcache_put (pop_data->bcache, h->data, 1)))
+    if (!(msg->fp = mutt_bcache_put (pop_data->bcache, cache_id (h->data), 1)))
     {
       /* no */
       bcache = 0;
@@ -619,7 +643,7 @@ int pop_fetch_message (MESSAGE* msg, CONTEXT* ctx, int msgno)
    * portion of the headers, those required for the main display.
    */
   if (bcache)
-    mutt_bcache_commit (pop_data->bcache, h->data);
+    mutt_bcache_commit (pop_data->bcache, cache_id (h->data));
   else
   {
     cache->index = h->index;
@@ -631,10 +655,12 @@ int pop_fetch_message (MESSAGE* msg, CONTEXT* ctx, int msgno)
   /* we replace envelop, key in subj_hash has to be updated as well */
   if (ctx->subj_hash && h->env->real_subj)
     hash_delete (ctx->subj_hash, h->env->real_subj, h, NULL);
+  mutt_label_hash_remove (ctx, h);
   mutt_free_envelope (&h->env);
   h->env = mutt_read_rfc822_header (msg->fp, h, 0, 0);
   if (ctx->subj_hash && h->env->real_subj)
-    hash_insert (ctx->subj_hash, h->env->real_subj, h, 1);
+    hash_insert (ctx->subj_hash, h->env->real_subj, h);
+  mutt_label_hash_add (ctx, h);
 
   h->data = uidl;
   h->lines = 0;
@@ -657,8 +683,13 @@ int pop_fetch_message (MESSAGE* msg, CONTEXT* ctx, int msgno)
   return 0;
 }
 
+static int pop_close_message (CONTEXT *ctx, MESSAGE *msg)
+{
+  return safe_fclose (&msg->fp);
+}
+
 /* update POP mailbox - delete messages from server */
-int pop_sync_mailbox (CONTEXT *ctx, int *index_hint)
+static int pop_sync_mailbox (CONTEXT *ctx, int *index_hint)
 {
   int i, j, ret = 0;
   char buf[LONG_STRING];
@@ -676,7 +707,7 @@ int pop_sync_mailbox (CONTEXT *ctx, int *index_hint)
       return -1;
 
     mutt_progress_init (&progress, _("Marking messages deleted..."),
-			M_PROGRESS_MSG, WriteInc, ctx->deleted);
+			MUTT_PROGRESS_MSG, WriteInc, ctx->deleted);
 
 #if USE_HCACHE
     hc = pop_hcache_open (pop_data, ctx->path);
@@ -692,7 +723,7 @@ int pop_sync_mailbox (CONTEXT *ctx, int *index_hint)
 	snprintf (buf, sizeof (buf), "DELE %d\r\n", ctx->hdrs[i]->refno);
 	if ((ret = pop_query (pop_data, buf, sizeof (buf))) == 0)
 	{
-	  mutt_bcache_del (pop_data->bcache, ctx->hdrs[i]->data);
+	  mutt_bcache_del (pop_data->bcache, cache_id (ctx->hdrs[i]->data));
 #if USE_HCACHE
 	  mutt_hcache_delete (hc, ctx->hdrs[i]->data, strlen);
 #endif
@@ -702,7 +733,7 @@ int pop_sync_mailbox (CONTEXT *ctx, int *index_hint)
 #if USE_HCACHE
       if (ctx->hdrs[i]->changed)
       {
-	mutt_hcache_store (hc, ctx->hdrs[i]->data, ctx->hdrs[i], 0, strlen, M_GENERATE_UIDVALIDITY);
+	mutt_hcache_store (hc, ctx->hdrs[i]->data, ctx->hdrs[i], 0, strlen, MUTT_GENERATE_UIDVALIDITY);
       }
 #endif
 
@@ -736,7 +767,7 @@ int pop_sync_mailbox (CONTEXT *ctx, int *index_hint)
 }
 
 /* Check for new messages and fetch headers */
-int pop_check_mailbox (CONTEXT *ctx, int *index_hint)
+static int pop_check_mailbox (CONTEXT *ctx, int *index_hint)
 {
   int ret;
   POP_DATA *pop_data = (POP_DATA *)ctx->data;
@@ -762,7 +793,7 @@ int pop_check_mailbox (CONTEXT *ctx, int *index_hint)
     return -1;
 
   if (ret > 0)
-    return M_NEW_MAIL;
+    return MUTT_NEW_MAIL;
 
   return 0;
 }
@@ -850,7 +881,7 @@ void pop_fetch_mail (void)
     goto finish;
   }
 
-  if (mx_open_mailbox (NONULL (Spoolfile), M_APPEND, &ctx) == NULL)
+  if (mx_open_mailbox (NONULL (Spoolfile), MUTT_APPEND, &ctx) == NULL)
     goto finish;
 
   delanswer = query_quadoption (OPT_POPDELETE, _("Delete messages from server?"));
@@ -860,7 +891,7 @@ void pop_fetch_mail (void)
 
   for (i = last + 1 ; i <= msgs ; i++)
   {
-    if ((msg = mx_open_new_message (&ctx, NULL, M_ADD_FROM)) == NULL)
+    if ((msg = mx_open_new_message (&ctx, NULL, MUTT_ADD_FROM)) == NULL)
       ret = -3;
     else
     {
@@ -875,10 +906,10 @@ void pop_fetch_mail (void)
 	ret = -3;
       }
 
-      mx_close_message (&msg);
+      mx_close_message (&ctx, &msg);
     }
 
-    if (ret == 0 && delanswer == M_YES)
+    if (ret == 0 && delanswer == MUTT_YES)
     {
       /* delete the message on the server */
       snprintf (buffer, sizeof (buffer), "DELE %d\r\n", i);
@@ -928,3 +959,15 @@ fail:
   mutt_socket_close (conn);
   FREE (&pop_data);
 }
+
+struct mx_ops mx_pop_ops = {
+  .open = pop_open_mailbox,
+  .open_append = NULL,
+  .close = pop_close_mailbox,
+  .open_msg = pop_fetch_message,
+  .close_msg = pop_close_message,
+  .check = pop_check_mailbox,
+  .commit_msg = NULL,
+  .open_new_msg = NULL,
+  .sync = pop_sync_mailbox,
+};
